import argparse

from irkit.plot.trecplot import pr_curve

import irkit.trec.results


def main():
    argparser = argparse.ArgumentParser()

    argparser.add_argument('--trec_results', help='trec_eval results files.',
                           required=True, type=argparse.FileType('r'), nargs='+')
    argparser.add_argument('--output', help='Name of the output file.', type=str,
                           default='output', required=False)

    args = argparser.parse_args()
    pr_curve([irkit.trec.results.load(f) for f in args.trec_results]).savefig('output',
                                                                              bbox_inches='tight')



"""
Functions and classes for dealing with English qrel files. The specification can be viewed at:
http://trec.nist.gov/data/qrels_eng/

>>> sample_qrels = '1 0 AP880212-0161 0\\n1 0 AP880216-0139 1\\n1 0 AP880216-0169 0'
>>> qrels = loads(sample_qrels)
>>> qrels.topic
['1', '1', '1']

>>> [x.topic for x in qrels['1']]
['1', '1', '1']

Harry Scells
Mar 2017
"""

import io

import os
from typing import List


class Qrel:
    """
    A line in a qrels file conforming to the specification at:
    http://trec.nist.gov/data/qrels_eng/
    """

    def __init__(self, topic: str, iteration: int, document_num: str, relevancy: int):
        self.topic = topic
        self.iteration = iteration
        self.document_num = document_num
        self.relevancy = relevancy

    def __str__(self):
        return '{} {} {} {}'.format(self.topic, self.iteration,
                                    self.document_num, self.relevancy)


class Qrels:
    """
    A python representation of a qrels file. It is a list of Qrel objects.
    """

    def __init__(self, qrels: List[Qrel]):
        self.qrels = qrels

    def __getattr__(self, field):
        """
        Access a column of the qrels by accessing the fields in a Qrel.
        :param field: One of the attributes (fields) of the qrel
        :return: The column for the field (i.e. only the topics, or only the document_num)
        """
        return [x.__getattribute__(field) for x in self.qrels]

    def __setitem__(self, key, value):
        raise Exception('Cannot set qrel values')

    def __delattr__(self, topic):
        self.qrels = [x for x in self.qrels if x.topic != topic]

    def __getitem__(self, topic):
        """
        Allow qrels to be indexed by topic id.
        :param topic: The topic
        :return: The rows of this topic
        """
        return [x for x in self.qrels if x.topic == topic]

    def __str__(self):
        return os.linesep.join([str(x) for x in self.qrels])

    def dumps(self):
        """
        Dump the qrels to a string.
        :return: Formatted qrels
        """
        return str(self)

    def dump(self, fp: io.IOBase):
        """
        Dump the qrels to a file
        :param fp: A File pointer
        """
        fp.writelines(self.dumps())


def loads(qrels: str) -> Qrels:
    """
    Load qrels from a string.
    :param qrels: Some string representation of qrels
    :return: Qrels object
    """
    data = []
    for line in qrels.split(os.linesep):
        if len(line.split()) > 0:
            topic, iteration, document_num, relevancy = line.split()
            data.append(Qrel(topic, int(iteration), document_num, int(relevancy)))
    return Qrels(data)


def load(qrels: io.IOBase) -> Qrels:
    """
    Load qrels from a file.
    :param qrels: File pointer
    :return: Qrels object
    """
    return loads(os.linesep.join(qrels.readlines()))



"""
Various plotting functions.

Harry Scells
Mar 2017
"""

import re

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from typing import List

from irkit.trec.results import TrecEvalResults

# This causes matplotlib to use Type 42 (a.k.a. TrueType) fonts for PostScript and PDF files.
# This allows you to avoid Type 3 fonts without limiting yourself to the stone-age technology
# of Type 1 fonts.
mpl.rcParams['pdf.fonttype'] = 42
mpl.rcParams['ps.fonttype'] = 42

plt.style.use('grayscale')
plt.style.use('seaborn-poster')
plt.style.use('seaborn-white')


def pr_curve(results: List[TrecEvalResults]) -> plt:
    """
    Create a precision-recall graph from trec_eval results.
    :param results: A list of TrecEvalResults files
    :return: a matplotlib plt object
    """

    names = [r.run_id for r in results]
    iprec = [[r.results['iprec_at_recall_0.00'],
              r.results['iprec_at_recall_0.10'],
              r.results['iprec_at_recall_0.20'],
              r.results['iprec_at_recall_0.30'],
              r.results['iprec_at_recall_0.40'],
              r.results['iprec_at_recall_0.50'],
              r.results['iprec_at_recall_0.60'],
              r.results['iprec_at_recall_0.70'],
              r.results['iprec_at_recall_0.80'],
              r.results['iprec_at_recall_0.90'],
              r.results['iprec_at_recall_1.00']] for r in results]

    recall = np.arange(0, 1.1, 0.1)

    mpl.rc('xtick', labelsize=35)
    mpl.rc('ytick', labelsize=35)

    plt.xlabel('Recall', fontsize=35)
    plt.ylabel('Interpolated Precision', fontsize=35)

    for p in iprec:
        plt.plot(recall, p, linewidth=10)

    plt.legend(names, fontsize=35)
    return plt


def topic_ap(results: List[TrecEvalResults], sort_on_ap=False):
    """

    :param results:
    :param sort_on_ap:
    :return:
    """

    def natural_sort(l):
        """

        :param l: List to sort
        :return: naturally sorted list
        """

        def convert(text):
            return int(text) if text.isdigit() else text.lower()

        def alphanum_key(key):
            return [convert(c) for c in re.split('([0-9]+)', key)]

        return sorted(l, key=alphanum_key)

    # noinspection PyShadowingNames
    # we really don't care about shadowing names in this case
    def plot(ap, ylabel, topic_names, sort_on_ap=False):
        """

        :param ap:
        :param ylabel:
        :param topic_names:
        :param sort_on_ap:
        :return:
        """
        if sort_on_ap:
            topic_names = [n for (d, n) in sorted(zip(ap, topic_names), reverse=True)]
            ap = sorted(ap, reverse=True)

        ind = np.arange(len(topic_names))
        width = 1.0
        plt.bar(ind, ap, width)
        plt.xticks(ind + 0.5 * width, topic_names, fontsize=10, rotation='vertical')
        plt.ylabel(ylabel)
        plt.xlabel('Topic')

        return plt

    if len(results) == 2:
        r1 = results[0]
        r2 = results[1]
        assert all([k in r2.queries for k in r1.queries.keys()]), 'Topic set is not the same!'
        assert all([k in r1.queries for k in r2.queries.keys()]), 'Topic set is not the same!'

        topic_names = natural_sort(r1.queries.keys())

        ap1 = np.array([r1.queries[q]['map'] for q in topic_names])
        ap2 = np.array([r2.queries[q]['map'] for q in topic_names])

        ap = ap1 - ap2
        ylabel = 'Average Precision difference'

        return plot(ap, ylabel, topic_names, sort_on_ap)
    elif len(results) == 1:
        r = results[0]
        topic_names = natural_sort(r.queries.keys())
        ap = np.array([r.queries[q]['map'] for q in topic_names])
        ylabel = 'Average Precision'

        return plot(ap, ylabel, topic_names, sort_on_ap)
    else:
        raise Exception('Can only plot either one or two TrecEvalResults objects.')



"""

Harry Scells
May 2017
"""


class TrecEvalRun(object):
    def __init__(self, topic: str, q: int, doc_id: str, rank: int, score: float, run_id: str):
        self.topic = topic
        self.q = q
        self.doc_id = doc_id
        self.rank = rank
        self.score = score
        self.run_id = run_id

    def __str__(self):
        return '{}\tQ{}\t{}\t{}\t{}\t{}'.format(self.topic, self.q, self.doc_id, self.rank,
                                                self.score, self.run_id)



"""
Functions and classes for dealing with trec_eval results files.

Harry Scells
Mar 2017
"""
import io
import os
from typing import Dict

eval_fields = {'num_q': int,
               'num_ret': int,
               'num_rel': int,
               'num_rel_ret': int,
               'map': float,
               'gm_map': float,
               'Rprec': float,
               'bpref': float,
               'recip_rank': float,
               'iprec_at_recall_0.00': float,
               'iprec_at_recall_0.10': float,
               'iprec_at_recall_0.20': float,
               'iprec_at_recall_0.30': float,
               'iprec_at_recall_0.40': float,
               'iprec_at_recall_0.50': float,
               'iprec_at_recall_0.60': float,
               'iprec_at_recall_0.70': float,
               'iprec_at_recall_0.80': float,
               'iprec_at_recall_0.90': float,
               'iprec_at_recall_1.00': float,
               'P_5': float,
               'P_10': float,
               'P_15': float,
               'P_20': float,
               'P_30': float,
               'P_100': float,
               'P_200': float,
               'P_500': float,
               'P_1000': float}


class TrecEvalResults:
    """
    An object that stores the results output by trec_eval.
    """

    def __init__(self, run_id: str, results: Dict, queries: Dict):
        self.run_id = run_id
        self.results = results
        self.queries = queries

    def __getitem__(self, query):
        """
        Allow trec results to be indexed by query num.
        :param query: The query to search
        :return: The rows of this topic
        """
        return self.queries[query]


def loads(trec_results: str) -> TrecEvalResults:
    """
    Load trec_eval results from a string.
    :param trec_results: Some string representation of trec results
    :return: TrecEvalResults object
    """
    run_id = ''
    results = {}
    queries = {}

    for line in trec_results.split(os.linesep):
        if not line.strip():
            continue
        field, query, value = line.split()
        if query == 'all':  # accumulated results over all queries
            if field == 'runid':
                run_id = value
            else:
                results[field] = value
        else:
            if query not in queries:
                queries[query] = {}
            queries[query][field] = float(value)

    return TrecEvalResults(run_id, results, queries)


def load(trec_result_file: io.IOBase) -> TrecEvalResults:
    """
    Load trec_eval results from a file.
    :param trec_result_file: File pointer
    :return: Qrels object
    """
    return loads(os.linesep.join(trec_result_file.readlines()))